'''Misc utilities for for multivariate normal distributions'''

######## Imports ########
import numpy as np
from scipy import stats
from gwalk.multivariate_normal import nparams_of_ndim
from gwalk.multivariate_normal.decomposition import cov_rescale
from gwalk.multivariate_normal import maha as mahalanobis_distance
from gwalk.multivariate_normal import mu_cov_of_params
from gwalk.multivariate_normal import cor_of_params
from gwalk.multivariate_normal import cov_of_params
from gwalk.multivariate_normal import std_of_cov

######## Declarations ########

__all__ = [
           "analytic_kl",
           "analytic_enclosed_integral",
           "analytic_kl_of_params",
           "cor_eigvals_by_eps",
           "random_gauss_params",
           "eigvals_satisfy_constraints",
           "U_of_cov",
           "det_eigh_of_cov",
          ]

######## Functions ########

## Functions for testing ##
def random_gauss_params(
                           ngauss,
                           ndim,
                           rs = None,
                          ):
    ''' Generate random parameters for Gaussians

    Parameters
    ----------
    ngauss : int
        Number of gaussians to generate
    ndim : int
        Dimensions of Gaussian
    rs : `~numpy.random.RandomState` (optional)
        Random state for number generation

    Returns
    -------
    X : `~numpy.ndarray` (ngauss, nparam)
        Random Gaussian parameters
    '''
    # Make sure you have a random state
    if rs is None:
        rs = np.random.RandomState()
    # Get the number of params
    nparams = nparams_of_ndim(ndim)
    # Pull from the uniform distribution :)
    X = rs.uniform(low=0.,high=1.,size=(nparams*ngauss)).reshape((ngauss, nparams))
    # Keep correlations manageable
    X[:,2*ndim + 1:] *= 0.5
    return X

## Conversions ##

def U_of_cov(cov, scale=None):
    ''' Compute the normalized eigenvectors for a covariance, rescaled
    
    Parameters
    ----------
    cov: `~numpy.ndarray` shape = (ngauss, ndim, ndim)
        Input array of covariance matrices
    scale: array like, shape = (ngauss, ndim)
        Input array of scale factors
    Returns
    -------
    U: `~numpy.ndarray` shape = (ngauss, ndim, ndim)
        Output eigenvectors
    '''
    # Rescale cov
    cov = cov_rescale(cov, scale)
    # Get the un-normalized eigenvalues and eigenvectors
    s, u = np.linalg.eigh(cov)
    # Get the scipy keep coefficient
    eps = stats._multivariate._eigvalsh_to_eps(s)
    # Pick your good eigenvalues
    keep = np.prod(s > eps, axis=1).astype(bool)
    # Initialize your d tesnsor
    d = np.zeros_like(s)
    # Assign log of kept s values
    d[keep] = np.log(s[keep])
    # Find the log of the determinant of the eigenvalues
    log_det_cov = np.sum(d, axis=1)
    # Compute the inverse square root of the eigenvalues
    s_pinv = np.zeros_like(s)
    s_pinv[keep] = 1./np.sqrt(s[keep])
    # Compute the final normalized eigenvalues
    U = u*s_pinv[:,None,:]
    # Return them
    return U

def det_eigh_of_cov(cov, scale=None):
    ''' Compute the log of the determinant of the eigenvalues
    
    Parameters
    ----------
    cov: `~numpy.ndarray` shape = (ngauss, ndim, ndim)
        Input array of covariance matrices
    scale: array like, shape = (ngauss, ndim)
        Input array of scale factors
    Returns
    -------
    U: `~numpy.ndarray` shape = (ngauss, ndim, ndim)
        Output eigenvectors
    '''
    # Rescale cov
    cov = cov_rescale(cov, scale)
    # Get the un-normalized eigenvalues and eigenvectors
    s, u = np.linalg.eigh(cov)
    # Get the scipy keep coefficient
    eps = stats._multivariate._eigvalsh_to_eps(s)
    # Pick your good eigenvalues
    keep = np.prod(s > eps, axis=1).astype(bool)
    # Initialize your d tesnsor
    d = np.zeros_like(s)
    # Assign log of kept s values
    d[keep] = s[keep]
    # Find the log of the determinant of the eigenvalues
    det = np.prod(d, axis=1)
    return det

def cor_eigvals_by_eps(X):
    ''' Get the product of the correlation eigenvalues divided by eps
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    Returns
    -------
    s: `~numpy.ndarray` shape = (ngauss,)
        Output product of raito of eigenvalues to eps
    '''
    # Get the correlation matrix
    cor = cor_of_params(X)
    # Get the eigenvalues of the correlation matrix
    cor_eigvals = np.linalg.eigvalsh(cor)
    ## Use scipy's eps value for consistency with resolving singular matrices ##
    eps = stats._multivariate._eigvalsh_to_eps(cor_eigvals)
    # Get the ratio of the eigenvalues to eps
    s = cor_eigvals / eps
    return s

def cov_eigvals_by_eps(X, dynamic_rescale=True):
    ''' Get the product of the covariance eigenvalues divided by eps
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    Returns
    -------
    s: `~numpy.ndarray` shape = (ngauss,)
        Output product of raito of eigenvalues to eps
    '''
    # Get the correlation matrix
    cov = cov_of_params(X)
    if dynamic_rescale:
        # Dynamic rescale
        std = std_of_cov(cov)
        scale = 1/std
        cov = cov_rescale(cov, scale)
    # Get the eigenvalues of the covariance matrix
    cov_eigvals = np.linalg.eigvalsh(cov)
    ## Use scipy's eps value for consistency with resolving singular matrices ##
    eps = stats._multivariate._eigvalsh_to_eps(cov_eigvals)
    # Get the ratio of the eigenvalues to eps
    s = cov_eigvals / eps
    return s



def eigvals_satisfy_constraints(X):
    ''' Find if correlation parameters satisfy constraints
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    Returns
    -------
    k: array like, shape = (ngauss,)
        Output boolean array of satisfy constraints
    '''
    # Get the correlation matrix
    cor = cor_of_params(X)
    # Get the eigenvalues of the correlation matrix
    cor_eigvals = np.linalg.eigvalsh(cor)
    ## Use scipy's eps value for consistency with resolving singular matrices ##
    eps = stats._multivariate._eigvalsh_to_eps(cor_eigvals)
    # Find the keep array
    k = np.prod(cor_eigvals > eps, axis=1).astype(bool)
    return k


def analytic_kl(mu1, cov1, mu2, cov2, scale1=None, scale2=None):
    '''Compute the analytic kl divergence between two Gaussians
    
    Parameters
    ----------
    mu1: array like, shape = (ngauss, ndim)
        Input Array of mu values
    cov1: array like, shape = (ngauss, ndim, ndim)
        Input Array of cov values
    mu2: array like, shape = (ngauss, ndim)
        Input Array of mu values
    cov2: array like, shape = (ngauss, ndim, ndim)
        Input Array of cov values
    scale1: array like, shape = (ngauss, ndim)
        Input array of scale factors
    scale2: array like, shape = (ngauss, ndim)
        Input array of scale factors

    Returns
    -------
    kl: array like, shape = (ngauss,)
        Output array of kl divergences
    '''
    from numpy.linalg import inv
    # Protect against usage
    if (len(mu1.shape) == 1) and (len(cov1.shape) == 2):
        mu1 = mu1[None,:]
        cov1 = cov1[None,:,:]
    elif (len(mu1.shape) != 2) or (len(cov1.shape) != 3):
        raise ValueError("mu or cov is not the right shape!")
    if (len(mu2.shape) == 1) and (len(cov2.shape) == 2):
        mu2 = mu2[None,:]
        cov2 = cov2[None,:,:]
    elif (len(mu2.shape) != 2) or (len(cov2.shape) != 3):
        raise ValueError("mu or cov is not the right shape!")
    # Identify useful information
    ngauss = mu1.shape[0]
    if not ngauss == 1:
        raise NotImplementedError
    ndim = mu1.shape[1]
    nparams = nparams_of_ndim(ndim)
    # Protect against usage
    if cov1.shape != (ngauss, ndim, ndim):
        raise ValueError("cov1 is not shape (ngauss, ndim, ndim)")
    if mu2.shape != (ngauss, ndim):
        raise ValueError("cov1 is not shape (ngauss, ndim, ndim)")
    if cov2.shape != (ngauss, ndim, ndim):
        raise ValueError("cov2 is not shape (ngauss, ndim, ndim)")

    ## Eigenvector Decomposition ##
    # Find eigenvalues and eigenvectors (see scipy)
    # Sum the determinant
    d1 = det_eigh_of_cov(cov1, scale=scale1)
    d2 = det_eigh_of_cov(cov2, scale=scale2)
    # Estimate the tensor U
    #U1 = U_of_eigh(s1, u1, k1)
    U2 = U_of_cov(cov2, scale=scale2)

    ## Calculate the mahalanobis factor ##
    #maha_21 = mahalanobis_cython(np.asarray(mu1,order='c'), mu2, U2, ngauss, 1, ndim)
    if not ((scale1 is None) and (scale2 is None)):
        scale = scale2/scale1
    elif scale1 is None:
        scale = scale2
    elif scale2 is None:
        scale = 1/scale1
    else:
        scale = np.ones(ndim)
    maha_21 = mahalanobis_distance(mu2, scale, U2, mu1)

    ## Calculate the trace of the inner product of cov2 with cov1
    icov2 = inv(cov_rescale(cov2,scale=scale2))
    trace_inner = np.trace(np.matmul(icov2, cov_rescale(cov1,scale=scale1)),axis1=-1,axis2=-2)

    ## Calculate the log ratio of determinants
    lrd = np.log(d1/d2)

    ## Put it all together
    KL = float(0.5* (maha_21 + trace_inner - lrd - ndim))
    return KL

#### Functions that assume normal model object ####
def analytic_kl_of_params(
                          X1,
                          X2,
                          scale1 = None,
                          scale2 = None,
                         ):
    '''Compute the analytic kl divergence between two Gaussians
    
    Parameters
    ----------
    X1: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    X2: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    scale1: array like, shape = (ngauss,ndim)
        Input scale for different parameter guesses
    scale2: array like, shape = (ngauss,ndim)
        Input scale for different parameter guesses
    '''
    X1_mu, X1_cov = mu_cov_of_params(X1)
    X2_mu, X2_cov = mu_cov_of_params(X2)
    return analytic_kl(X1_mu, X1_cov, X2_mu, X2_cov, scale1=scale1, scale2=scale2)

def analytic_enclosed_integral(X, limits, scale):
    '''Compute the enclosed integral in a bounded interval
    
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    limits: array like, shape = (ndim, 2)
        Input Array of values
    scale: array like, shape = (ngauss,ndim)
        Input scale for different parameter guesses
    '''
    from scipy import stats
    from gwalk.multivariate_normal.decomposition import mu_of_params
    from gwalk.multivariate_normal.decomposition import std_of_params
    from gwalk.multivariate_normal.decomposition import params_rescale
    # Protect against single set of parameters
    X = params_rescale(X, scale=scale)
    mu  = mu_of_params(X)
    sig = std_of_params(X)
    # Extract information
    ngauss, ndim = mu.shape[0], mu.shape[1]
    # Protect against single set of limits
    if len(limits.shape) == 2:
        limits = limits[None,:,:]
    # Initialize integrated sum
    integral = np.ones(ngauss)
    # Loop the gaussians
    for i in range(ngauss):
        # Loop the dimensions
        for j in range(ndim):
            # Check limits
            if limits.shape[0] == 1:
                limit_indx = 0
            else:
                limit_indx = i
            # Find lower bound
            low  = stats.norm.cdf(limits[limit_indx,j,0],loc=mu[i,j],scale=sig[i,j])
            high = stats.norm.cdf(limits[limit_indx,j,1],loc=mu[i,j],scale=sig[i,j])
            # Update integral
            integral[i] *= high - low
    return integral





from gwalk.multivariate_normal import MultivariateNormal
#from gwalk.bounded_multivariate_normal import MultivariateNormal
#from .NormalApproximateLikelihood import NAL
#from .NAL import NAL
#from .catalog import Catalog
#from .data import database
#from .catalog import coordinates
#from .grid import Grid



#!/home/xevra/.local/bin/python3
'''\
Generate the likelihood function of each event
'''

######## Imports ########
from os.path import join, isfile, isdir
import os
import numpy as np
from .percentile_levels import percentile_levels
from ..utils.multivariate_normal import multivariate_normal_marginal1d
from ..utils.multivariate_normal import multivariate_normal_marginal2d

######## Plot Tools ########

def likelihood_1d(
                  ax,
                  Norm,
                  mesh,
                  index,
                  labels,
                  **kwargs
                 ):
    '''\
    make a 1d pdf for an ND gaussian
    
    Inputs:
        ax:     A plot axis
        Norm:   A MultivariateNormal object from basil.fit.models
        ind:    The index of the dimension we are plotting
        limits: The soft limits (that describe our distribution)
        bins:   The number of bins used on the associated histogram
    '''

    # Initialize limits
    xlims = mesh.limits[index]
    # Initialize dimensionality
    ndim = mesh.ndim

    # Generate x_test and y_test
    y_mesh, L_mesh = mesh.fetch_1d_evaluations(index)

    # Generate y_norm
    L_norm = multivariate_normal_marginal1d(
            Norm.check_sample(Norm.read_guess()),
            y_mesh,
            index,
            scale=Norm.scale,
           )

    # Normalize L_norm
    L_norm /= np.sum(L_norm)

    y_mesh = y_mesh.flatten()
    L_norm = L_norm.flatten()

    # Plot marginalized pdf of Normal object
    ax.plot(y_mesh, L_mesh, label = labels[0], **kwargs)
    ax.plot(y_mesh, L_norm, label = labels[1], **kwargs)
    ax.set_xlim(xlims)

def likelihood_2d(
                  ax,
                  Norm,
                  mesh,
                  index, jndex,
                  cmap = 'magma',
                  levels = [0.25, 0.5, 0.75],
                  **kwargs
                 ):
    '''\
    Plot some Gaussian contours!
    '''
    #from scipy.signal import convolve
    
    # Update limits
    xlims, ylims = mesh.limits[jndex], mesh.limits[index]
    # Initialize dimensionality
    ndim = mesh.ndim
    # Find 2d Evaluations 
    y_mesh, L_mesh = mesh.fetch_2d_evaluations(index, jndex)
    # Find L_norm
    L_norm = multivariate_normal_marginal2d(
            Norm.check_sample(Norm.read_guess()),
            y_mesh,
            index, jndex,
            scale=Norm.scale,
           )
    # Normalize L_norm
    L_norm /= np.sum(L_norm)
    # Recall resolution
    res = mesh.evaluation_res

    # Identify xgrid and ygrid
    x_grid = y_mesh[:,1].reshape((res,res))
    y_grid = y_mesh[:,0].reshape((res,res))
    # Identify xspace and yspace
    x_space = x_grid[0,:]
    y_space = y_grid[:,0]
    # Reshape outputs
    L_mesh = L_mesh.reshape((res,res))
    L_norm = L_norm.reshape((res,res))

    # Generate contour levels for catalog
    levels_mesh = percentile_levels(x_grid, y_grid, L_mesh, levels)

    # Generate contour levels for normal
    levels_norm = percentile_levels(x_grid, y_grid, L_norm, levels)

    # Plot contours for catalog
    ax.contour(
               x_space,
               y_space,
               L_mesh,
               levels = levels_mesh,
               cmap = cmap,
               vmin = 0.,
               vmax = np.max(L_mesh),
               linestyles = 'solid',
               **kwargs
              )
       
    # Plot contours for normal
    ax.contour(
               x_space,
               y_space,
               L_norm,
               levels = levels_norm,
               cmap = cmap,
               vmin = 0.,
               vmax = np.max(L_norm),
               linestyles = 'dotted',
               **kwargs
              )
       
    # Set limits
    ax.set_xlim(xlims)
    ax.set_ylim(ylims)

    
                     

######## Plots ########

#### Corner Plots ####

def corner_cross_sample_normal(
                               savename,
                               Norm,
                               mesh,
                               levels = [0.25, 0.5, 0.75],
                               title=None,
                               evaluation_res=100,
                               sample_labels=[None,None],
                               scale=1.,
                               rot_xlabel=False,
                               tight=False,
                               legend_loc=[0.45,0.68],
                               linewidth_scale=1.,
                              ):
    '''\
    Generate a corner plot for posterior/likelihood samples with model

    Inputs:
        title: title for plot
        bins: histogram/pdf bins
    '''
    #### Imports ####
    # Matplotlib 
    import matplotlib
    matplotlib.use("Agg")
    from matplotlib import pyplot as plt
    # Public

    #### Initialize Environment ####

    # Find the dimensionality
    ndim = mesh.ndim

    # Load data
    mesh.generate_evaluation_set(evaluation_res)

    # Initialize style
    plt.style.use('bmh')
    ## Initialize axes
    fig, axes = plt.subplots(
                             nrows = ndim,
                             ncols = ndim,
                             figsize = (3.375*scale,3.375*scale),
                             sharex = 'col',
                            )

    # Initialize ticks
    all_ticks = []

    #### 1-d plots ####
    for i in range(ndim):
        ## axes[i,i] ##
        # Here, we want to plot a 1-d histogram of the posterior samples
        # alongside a histogram of the reweighted likelihood samples resampled
        # from our gaussian
        axes[i,i].tick_params(axis='both',which='both',labelsize=6*scale)
        if rot_xlabel:
            axes[i,i].tick_params(axis='x',which='both',labelrotation=90,pad=0)
        if i == 0:
            labels = sample_labels
        else:
            labels = np.empty_like(sample_labels,dtype=object)
            for j in range(len(labels)):
                labels[j] = None
        
        # Make the appropriate plots
        likelihood_1d(
                      axes[i,i],
                      Norm,
                      mesh,
                      i,
                      labels,
                      linewidth=1.0*linewidth_scale,
                     )

        # Remove yticklabels
        axes[i,i].set_yticklabels([])
        # Append ticks
        all_ticks.append(axes[i,i].get_xticks())
        # Handle axis labels
        axes[-1,i].set_xlabel(Norm._parameters["mu_%d"%i].label,size=10*scale)

    #### 2-d plots ####
    for i in range(ndim):
        for j in range(i):
            # Remove unwanted plot
            axes[j,i].remove()

            ## axes[i,j] ##
            # Here we want a 2D histogram of the posterior samples alongside 
            # The reweighted likelihood samples resampled from our gaussian
            ax = axes[i,j]
            ax.tick_params(axis='both',which='both',labelsize=6*scale)
            if rot_xlabel:
                ax.tick_params(axis='x',which='both',labelrotation=90,pad=0)
    
            # Make the appropriate plots
            likelihood_2d(
                          ax,
                          Norm,
                          mesh,
                          i,
                          j,
                          linewidths = 1.5*linewidth_scale,
                         )

            if j != 0:
                # Remove yticklabels
                ax.set_yticklabels([])

    # Fill the legend for the contours
    axes[0,0].plot(
                   [], [], 
                   color = 'gray', 
                   linestyle = 'solid',
                   label = sample_labels[0],
                  )
    axes[0,0].plot(
                   [], [], 
                   color = 'gray', 
                   linestyle = 'dotted',
                   label = sample_labels[1],
                  )

    # Make 2D legend
    fig.legend(loc=legend_loc, prop={"size":8*scale})
    ## Wrapping up
    if not (title is None):
        fig.suptitle(title, fontsize = float(12*scale))
    if tight:
        plt.tight_layout()

    # Make sure there's a place to store the figures #
    plt.savefig(savename)
    plt.close()



''' Prior functions from ROS, from some version of RIFT
'''

# mcmin, mcmax : to be defined later
def mc_prior(mc, mc_min, mc_max):
    ''' An uninformative mass prior
    not normalized; see section II.C of https://arxiv.org/pdf/1701.01137.pdf

    Parameters
    ----------
    mc: array like, shape = (npts)
        Input mc values
    mc_min: float
        Input minimum mc value
    mc_max: float
        Input maximum mc value
    '''
    return 2*mc/(mc_max**2-mc_min**2)

def q_prior(x):
    ''' An uninformative mass ratio prior
    # not normalized; see section II.C of https://arxiv.org/pdf/1701.01137.pdf

    Parameters
    ----------
    q: array like, shape = (npts)
        Input mass ratio values
    '''
    return 1./(1+x)**2

def m1_prior(m1):
    ''' An uninformative prior in m1
    Parameters
    ----------
    m1: array_like, shape = (npts)
        Input m1 values
    '''
    return 1./200

def m2_prior(m2):
    ''' An uninformative prior in m2
    Parameters
    ----------
    m2: array_like, shape = (npts)
        Input m2 values
    '''
    return 1./200

def s1z_prior(s1z, chi_max = 1.):
    ''' An uninformative prior in s1z
    Parameters
    ----------
    s1z: array_like, shape = (npts)
        Input s1z values
    chi_max: float, optional
        Input maximum allowed spin
    '''
    return 1./(2*chi_max)

def s2z_prior(s2z, chi_max = 1.):
    ''' An uninformative prior in s2z
    Parameters
    ----------
    s2z: array_like, shape = (npts)
        Input s2z values
    chi_max: float, optional
        Input maximum allowed spin
    '''
    return 1./(2*chi_max)

def unscaled_eta_prior_cdf(eta_val):
    """ Uninformative eta prior cdf value

    cumulative for integration of x^(-6/5)(1-4x)^(-1/2) from eta_min to 1/4.
    Used to normalize the eta prior
    Derivation in mathematica:
       Integrate[ 1/\[Eta]^(6/5) 1/Sqrt[1 - 4 \[Eta]], {\[Eta], \[Eta]min, 1/4}]

    Parameters
    ----------
    eta_val: float
        Input given value of eta
    """
    import numpy as np
    import scipy.special
    return  2**(2./5.) * \
                np.sqrt(np.pi) * \
                (
                 scipy.special.gamma(-0.2) / \
                 scipy.special.gamma(0.3) 
                ) \
                + 5 * (
                       scipy.special.hyp2f1(-0.2,0.5,0.8, 4*eta_val) / \
                       (eta_val**(0.2))
                      )

def eta_prior(eta, norm_factor=1.44):
    """ Uninformative eta prior

    Change norm_factor by the output 
    Parameters
    ----------

    eta: array_like, shape = (npts)
        Input eta values
    norm_factor: float, optional
        Input normalization factor
    """
    import numpy as np
    return (np.power(eta, -6./5.) * np.power(1-4.*eta, -0.5)) / norm_factor

def delta_mc_prior(x,norm_factor=1.44):
    """ Uninformative delta mc prior
    delta_mc = sqrt(1-4eta)  <-> eta = 1/4(1-delta^2)
    Transform the prior above
    Parameters
    ----------
    x: array_like, shape = (npts)
        Input sample values
    norm_factor: float, optional
        Input normalization factor
    """
    import numpy as np
    eta_here = 0.25*(1 -x*x)
    return 2.*np.power(eta_here, -6./5.)/norm_factor

def m_prior(x):
    '''Uninformative m prior

    Parameters
    ----------
    x: array_like, shape = (npts)
        Input sample values
    '''
    # uniform in mass, use a square.
    #Should always be used as m1,m2 in pairs.
    #Note this does NOT restrict m1>m2.
    return 1/(1e3-1.)

def eta_norm_factor(eta_min, eta_max):
    '''normalization factor for uninformative eta prior
    Parameters
    ----------
    eta_min: float
        Input minimum value for eta
    eta_max: float
        Input maximum value for eta
    '''
    norm_factor = unscaled_eta_prior_cdf(eta_min) - \
                    unscaled_eta_prior_cdf(eta_max)
    return norm_factor

def mass_area(mc_min, mc_max, eta_min, eta_max):
    ''' Normalization factor for mass prior
    Parameters
    ----------
    mc_min: float
        Input minimum value for mc
    mc_max: float
        Input maximum value for mc
    eta_min: float
        Input minimum value for eta
    eta_max: float
        Input maximum value for eta
    '''
    # mass normalization 
    # (assuming mc, eta limits are bounds - as is invariably the case)
    mass_area = 0.5* \
                    (mc_max**2 - mc_min**2)* \
                    (
                     unscaled_eta_prior_cdf(eta_min) -\
                     unscaled_eta_prior_cdf(eta_max)
                    )
    return mass_area




import numpy as np
from scipy.stats import gaussian_kde
from scipy.special import spence as PL

def Di(z):

    """
    Wrapper for the scipy implmentation of Spence's function.
    Note that we adhere to the Mathematica convention as detailed in:
    https://reference.wolfram.com/language/ref/PolyLog.html

    Inputs
    z: A (possibly complex) scalar or array

    Returns
    Array equivalent to PolyLog[2,z], as defined by Mathematica
    """

    return PL(1.-z+0j)

def chi_effective_prior_of_aligned_spins(q,aMax,xs):

    """
    Function defining the conditional priors p(chi_eff|q) corresponding to
    uniform, aligned component spin priors.

    Inputs
    q: Mass ratio value (according to the convention q<1)
    aMax: Maximum allowed dimensionless component spin magnitude
    xs: Chi_effective value or values at which we wish to compute prior

    Returns:
    Array of prior values
    """

    # Ensure that `xs` is an array and take absolute value
    xs = np.reshape(xs,-1)

    # Set up various piecewise cases
    pdfs = np.zeros(xs.size)
    caseA = (xs>aMax*(1.-q)/(1.+q))*(xs<=aMax)
    caseB = (xs<-aMax*(1.-q)/(1.+q))*(xs>=-aMax)
    caseC = (xs>=-aMax*(1.-q)/(1.+q))*(xs<=aMax*(1.-q)/(1.+q))

    # Select relevant effective spins
    x_A = xs[caseA]
    x_B = xs[caseB]
    x_C = xs[caseC]

    if np.sum(caseA) > 0:
        pdfs[caseA] = (1.+q[caseA])**2.*(aMax-x_A)/(4.*q[caseA]*aMax**2)
    if np.sum(caseB) > 0:
        pdfs[caseB] = (1.+q[caseB])**2.*(aMax+x_B)/(4.*q[caseB]*aMax**2)
    if np.sum(caseC) > 0:
        pdfs[caseC] = (1.+q[caseC])/(2.*aMax)

    return pdfs

def chi_effective_prior_of_isotropic_spins(q,aMax,xs):

    """
    Function defining the conditional priors p(chi_eff|q) corresponding to
    uniform, isotropic component spin priors.

    Inputs
    q: Mass ratio value (according to the convention q<1)
    aMax: Maximum allowed dimensionless component spin magnitude
    xs: Chi_effective value or values at which we wish to compute prior

    Returns:
    Array of prior values
    """

    # Ensure that `xs` is an array and take absolute value
    xs = np.reshape(np.abs(xs),-1)

    # Set up various piecewise cases
    pdfs = np.ones(xs.size,dtype=complex)*(-1.)
    caseZ = (xs==0)
    caseA = (xs>0)*(xs<aMax*(1.-q)/(1.+q))*(xs<q*aMax/(1.+q))
    caseB = (xs<aMax*(1.-q)/(1.+q))*(xs>q*aMax/(1.+q))
    caseC = (xs>aMax*(1.-q)/(1.+q))*(xs<q*aMax/(1.+q))
    caseD = (xs>aMax*(1.-q)/(1.+q))*(xs<aMax/(1.+q))*(xs>=q*aMax/(1.+q))
    caseE = (xs>aMax*(1.-q)/(1.+q))*(xs>aMax/(1.+q))*(xs<aMax)
    caseF = (xs>=aMax)

    # Select relevant effective spins
    x_A = xs[caseA]
    x_B = xs[caseB]
    x_C = xs[caseC]
    x_D = xs[caseD]
    x_E = xs[caseE]

    pdfs[caseZ] = (1.+q)/(2.*aMax)*(2.-np.log(q))

    pdfs[caseA] = (1.+q)/(4.*q*aMax**2)*(
                    q*aMax*(4.+2.*np.log(aMax) - np.log(q**2*aMax**2 - (1.+q)**2*x_A**2))
                    - 2.*(1.+q)*x_A*np.arctanh((1.+q)*x_A/(q*aMax))
                    + (1.+q)*x_A*(Di(-q*aMax/((1.+q)*x_A)) - Di(q*aMax/((1.+q)*x_A)))
                    )

    pdfs[caseB] = (1.+q)/(4.*q*aMax**2)*(
                    4.*q*aMax
                    + 2.*q*aMax*np.log(aMax)
                    - 2.*(1.+q)*x_B*np.arctanh(q*aMax/((1.+q)*x_B))
                    - q*aMax*np.log((1.+q)**2*x_B**2 - q**2*aMax**2)
                    + (1.+q)*x_B*(Di(-q*aMax/((1.+q)*x_B)) - Di(q*aMax/((1.+q)*x_B)))
                    )

    pdfs[caseC] = (1.+q)/(4.*q*aMax**2)*(
                    2.*(1.+q)*(aMax-x_C)
                    - (1.+q)*x_C*np.log(aMax)**2.
                    + (aMax + (1.+q)*x_C*np.log((1.+q)*x_C))*np.log(q*aMax/(aMax-(1.+q)*x_C))
                    - (1.+q)*x_C*np.log(aMax)*(2. + np.log(q) - np.log(aMax-(1.+q)*x_C))
                    + q*aMax*np.log(aMax/(q*aMax-(1.+q)*x_C))
                    + (1.+q)*x_C*np.log((aMax-(1.+q)*x_C)*(q*aMax-(1.+q)*x_C)/q)
                    + (1.+q)*x_C*(Di(1.-aMax/((1.+q)*x_C)) - Di(q*aMax/((1.+q)*x_C)))
                    )

    pdfs[caseD] = (1.+q)/(4.*q*aMax**2)*(
                    -x_D*np.log(aMax)**2
                    + 2.*(1.+q)*(aMax-x_D)
                    + q*aMax*np.log(aMax/((1.+q)*x_D-q*aMax))
                    + aMax*np.log(q*aMax/(aMax-(1.+q)*x_D))
                    - x_D*np.log(aMax)*(2.*(1.+q) - np.log((1.+q)*x_D) - q*np.log((1.+q)*x_D/aMax))
                    + (1.+q)*x_D*np.log((-q*aMax+(1.+q)*x_D)*(aMax-(1.+q)*x_D)/q)
                    + (1.+q)*x_D*np.log(aMax/((1.+q)*x_D))*np.log((aMax-(1.+q)*x_D)/q)
                    + (1.+q)*x_D*(Di(1.-aMax/((1.+q)*x_D)) - Di(q*aMax/((1.+q)*x_D)))
                    )

    pdfs[caseE] = (1.+q)/(4.*q*aMax**2)*(
                    2.*(1.+q)*(aMax-x_E)
                    - (1.+q)*x_E*np.log(aMax)**2
                    + np.log(aMax)*(
                        aMax
                        -2.*(1.+q)*x_E
                        -(1.+q)*x_E*np.log(q/((1.+q)*x_E-aMax))
                        )
                    - aMax*np.log(((1.+q)*x_E-aMax)/q)
                    + (1.+q)*x_E*np.log(((1.+q)*x_E-aMax)*((1.+q)*x_E-q*aMax)/q)
                    + (1.+q)*x_E*np.log((1.+q)*x_E)*np.log(q*aMax/((1.+q)*x_E-aMax))
                    - q*aMax*np.log(((1.+q)*x_E-q*aMax)/aMax)
                    + (1.+q)*x_E*(Di(1.-aMax/((1.+q)*x_E)) - Di(q*aMax/((1.+q)*x_E)))
                    )

    pdfs[caseF] = 0.

    # Deal with spins on the boundary between cases
    if np.any(pdfs==-1):
        boundary = (pdfs==-1)
        pdfs[boundary] = 0.5*(chi_effective_prior_of_isotropic_spins(q,aMax,xs[boundary]+1e-6)\
                        + chi_effective_prior_of_isotropic_spins(q,aMax,xs[boundary]-1e-6))

    return np.real(pdfs)

def chi_p_prior_of_isotropic_spins(q,aMax,xs):

    """
    Function defining the conditional priors p(chi_p|q) corresponding to
    uniform, isotropic component spin priors.

    Inputs
    q: Mass ratio value (according to the convention q<1)
    aMax: Maximum allowed dimensionless component spin magnitude
    xs: Chi_p value or values at which we wish to compute prior

    Returns:
    Array of prior values
    """

    # Ensure that `xs` is an array and take absolute value
    xs = np.reshape(xs,-1)

    # Set up various piecewise cases
    pdfs = np.zeros(xs.size)
    caseA = xs<q*aMax*(3.+4.*q)/(4.+3.*q)
    caseB = (xs>=q*aMax*(3.+4.*q)/(4.+3.*q))*(xs<aMax)

    # Select relevant effective spins
    x_A = xs[caseA]
    x_B = xs[caseB]

    pdfs[caseA] = (1./(aMax**2*q))*((4.+3.*q)/(3.+4.*q))*(
                    np.arccos((4.+3.*q)*x_A/((3.+4.*q)*q*aMax))*(
                        aMax
                        - np.sqrt(aMax**2-x_A**2)
                        + x_A*np.arccos(x_A/aMax)
                        )
                    + np.arccos(x_A/aMax)*(
                        aMax*q*(3.+4.*q)/(4.+3.*q)
                        - np.sqrt(aMax**2*q**2*((3.+4.*q)/(4.+3.*q))**2 - x_A**2)
                        + x_A*np.arccos((4.+3.*q)*x_A/((3.+4.*q)*aMax*q))
                        )
                    )
                    
    pdfs[caseB] = (1./aMax)*np.arccos(x_B/aMax)

    return pdfs

def joint_prior_of_isotropic_spins(q,aMax,xeffs,xps,ndraws=10000,bw_method='scott'):

    """
    Function to calculate the conditional priors p(xp|xeff,q) on a set of {xp,xeff,q} posterior samples.

    INPUTS
    q: Mass ratio
    aMax: Maximimum spin magnitude considered
    xeffs: Effective inspiral spin samples
    xps: Effective precessing spin values
    ndraws: Number of draws from the component spin priors used in numerically building interpolant

    RETURNS
    p_chi_p: Array of priors on xp, conditioned on given effective inspiral spins and mass ratios
    """

    # Convert to arrays for safety
    xeffs = np.reshape(xeffs,-1)
    xps = np.reshape(xps,-1)
    
    # Compute marginal prior on xeff, conditional prior on xp, and multiply to get joint prior!
    p_chi_eff = chi_effective_prior_of_isotropic_spins(q,aMax,xeffs)
    p_chi_p_given_chi_eff = np.array([chi_p_prior_given_chi_eff_q(q,aMax,xeffs[i],xps[i],ndraws,bw_method) for i in range(len(xeffs))])
    joint_p_chi_p_chi_eff = p_chi_eff*p_chi_p_given_chi_eff

    return joint_p_chi_p_chi_eff

def chi_p_prior_given_chi_eff_q(q,aMax,xeff,xp,ndraws=10000,bw_method='scott'):

    """
    Function to calculate the conditional prior p(xp|xeff,q) on a single {xp,xeff,q} posterior sample.
    Called by `joint_prior_of_isotropic_spins`.

    INPUTS
    q: Single posterior mass ratio sample
    aMax: Maximimum spin magnitude considered
    xeff: Single effective inspiral spin sample
    xp: Single effective precessing spin value
    ndraws: Number of draws from the component spin priors used in numerically building interpolant

    RETURNS
    p_chi_p: Prior on xp, conditioned on given effective inspiral spin and mass ratio
    """

    # Draw random spin magnitudes.
    # Note that, given a fixed chi_eff, a1 can be no larger than (1+q)*chi_eff,
    # and a2 can be no larger than (1+q)*chi_eff/q
    a1 = np.random.random(ndraws)*aMax
    a2 = np.random.random(ndraws)*aMax

    # Draw random tilts for spin 2
    cost2 = 2.*np.random.random(ndraws)-1.

    # Finally, given our conditional value for chi_eff, we can solve for cost1
    # Note, though, that we still must require that the implied value of cost1 be *physical*
    cost1 = (xeff*(1.+q) - q*a2*cost2)/a1  

    # While any cost1 values remain unphysical, redraw a1, a2, and cost2, and recompute
    # Repeat as necessary
    while np.any(cost1<-1) or np.any(cost1>1):   
        to_replace = np.where((cost1<-1) | (cost1>1))[0]   
        a1[to_replace] = np.random.random(to_replace.size)*aMax
        a2[to_replace] = np.random.random(to_replace.size)*aMax
        cost2[to_replace] = 2.*np.random.random(to_replace.size)-1.    
        cost1 = (xeff*(1.+q) - q*a2*cost2)/a1   
            
    # Compute precessing spins and corresponding weights, build KDE
    # See `Joint-ChiEff-ChiP-Prior.ipynb` for a discussion of these weights
    Xp_draws = chi_p_of_components(a1,a2,cost1,cost2,q)
    jacobian_weights = (1.+q)/a1
    prior_kde = gaussian_kde(Xp_draws,weights=jacobian_weights,bw_method=bw_method)

    # Compute maximum chi_p
    if (1.+q)*np.abs(xeff)/q<aMax:
        max_Xp = aMax
    else:
        max_Xp = np.sqrt(aMax**2 - ((1.+q)*np.abs(xeff)-q)**2.)

    # Set up a grid slightly inside (0,max chi_p) and evaluate KDE
    reference_grid = np.linspace(0.05*max_Xp,0.95*max_Xp,30)
    reference_vals = prior_kde(reference_grid)

    # Manually prepend/append zeros at the boundaries
    reference_grid = np.concatenate([[0],reference_grid,[max_Xp]])
    reference_vals = np.concatenate([[0],reference_vals,[0]])
    norm_constant = np.trapz(reference_vals,reference_grid)

    # Interpolate!
    p_chi_p = np.interp(xp,reference_grid,reference_vals)
    return p_chi_p

def chi_p_of_components(a1,a2,cost1,cost2,q):

    """
    Helper function to define effective precessing spin parameter from component spins

    INPUTS
    a1: Primary dimensionless spin magnitude
    a2: Secondary's spin magnitude
    cost1: Cosine of the primary's spin-orbit tilt angle
    cost2: Cosine of the secondary's spin-orbit tilt
    q: Mass ratio

    RETRUNS
    chi_p: Corresponding precessing spin value
    """

    sint1 = np.sqrt(1.-cost1**2)
    sint2 = np.sqrt(1.-cost2**2)
    
    return np.maximum(a1*sint1,((3.+4.*q)/(4.+3.*q))*q*a2*sint2)



'''\
Percentile based levels for 2D plots
'''
#https://stackoverflow.com/questions/37890550/python-plotting-percentile-contour-lines-of-a-probability-distribution

def percentile_levels(x,y,z, levels, nbins = 1000):
    '''\
    Generate percentile levels

    x,y,z should already be a meshgrid
    '''
    import numpy as np
    from scipy import interpolate

    # Define space for integral
    t = np.linspace(0, z.max(), nbins)

    # Create a mask for the integral
    mask = z <= t[:, None, None]

    # Perform the integral
    integral = (mask * z).sum(axis = (1,2))

    # Interpolate
    f = interpolate.interp1d(integral, t)
    # Return the function evaluated on the levels
    return f(levels)

def bivariate_normal(x, y, xscale, yscale, xloc, yloc):
    import numpy as np
    from scipy.stats import multivariate_normal
    mean = [xloc, yloc]
    cov = [[xscale, 0], [0, yscale]]
    rv = multivariate_normal(mean, cov)
    X = np.stack((x,y)).T
    return rv.pdf(X)

def example():
    import numpy as np
    import matplotlib
    import matplotlib.pyplot as plt
    x = np.linspace(-3,3,100)
    y = np.linspace(-3,3,100)
    X, Y = np.meshgrid(x, y)

    #z1 = bivariate_normal(X, Y, .5, .5, 0., 0.)
    z2 = bivariate_normal(X, Y, .4, .4, .5, .5)
    z3 = bivariate_normal(X, Y, .6, .2, -1.5, 0.)
    z = z3 + z2
    #z = z1 + z2 + z3
    z = z / z.sum()

    t_contours = percentile_levels(X,Y,z,np.arange(0.,1.,0.1))
    plt.imshow(z.T, origin='lower', extent=[-3,3,-3,3], cmap="gray")
    plt.contour(z.T, t_contours, extent=[-3,3,-3,3])
    plt.show()




#!/usr/bin/env python3
'''\
Test the Mesh object fitting, saving, and loading
'''
######## IMPORTS ########
import numpy as np

######## Settings ########

N_RVS = int(1e6)
N_MAX = 3
HYPERCUBE_RES = 100
N_SELECT = 1000
SEED = 0
RS = np.random.RandomState(SEED)
NDIM = 3
NSELECT = 1000
LIMITS = np.asarray([
                     [-7., 7.],
                     [-7., 7.],
                     [-7., 7.],
                    ])


######## Tests ########
def test_optimize_grid():
    #### Imports ####
    import time
    import numpy as np
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.optimize_likelihood_grid import optimize_likelihood_grid
    from gwalk.multivariate_normal import pdf as multivariate_normal_pdf
    from gwalk.multivariate_normal import params_of_offset_mu_cov
    from gwalk.multivariate_normal import mu_cov_of_params
    from gwalk.multivariate_normal import cor_of_cov
    from gwalk.multivariate_normal import offset_of_params
    from gwalk.multivariate_normal import MultivariateNormal

    #### Generate a grid ####
    RS.seed(SEED)
    X0 = random_gauss_params(1, NDIM, rs=RS)
    mu, cov = mu_cov_of_params(X0)
    mu, cov = mu_cov_of_params(X0)
    X0 = X0.flatten()
    #objective = "kl"
    objective = "lstsq"
    method="SLSQP"
    #method="trust-constr"

    #### run code ####

    ## random samples ##
    samples = (RS.uniform(size=(N_RVS,NDIM))*(LIMITS[:,1]-LIMITS[:,0])) + LIMITS[:,0]
    ## pdf values ##
    values = multivariate_normal_pdf(
                                     mu,
                                     cov,
                                     samples,
                                     log_scale=True,
                                    ).flatten() + offset_of_params(X0)
    ## Generate from samples ##
    MVsimple = MultivariateNormal.from_samples(samples, seed=SEED)

    ## TODO KEEP ##

    ## Downselect samples ##
    select = np.argsort(values)[-NSELECT:]
    samples = samples[select]
    values = values[select]

    ## Fit grid ##
    t0 = time.time()
    MV = optimize_likelihood_grid(
                                  samples,
                                  lnP=values,
                                  seed=SEED,
                                  objective=objective,
                                  method=method,
                                 )
    t1 = time.time()


    Xg = MV.params
    ## Print parameter values ##
    print("fit time: %f seconds"%(t1-t0))
    print("Objective: %s"%(objective))
    print("--------")
    print("True params:")
    print(X0)
    #TODO
    #print("kl divergence with true params:")
    #print(grid.nal_kl_div(MV,params))
    print("--------")
    print("Simple params:")
    print(MVsimple.params)
    #print("kl divergence with simple params:")
    #print(grid.nal_kl_div(MV,grid.X_simple))
    print("ln(simple error:)")
    print(np.log(np.abs(MVsimple.params - X0)))
    print("--------")
    print("Optimized params")
    print(Xg)
    #print("kl divergence with optimized params:")
    #print(grid.nal_kl_div(MV,Xg))
    print("ln(optimized error:)")
    print(np.log(np.abs(Xg - X0)))
    return

######## Main ########
def main():
    test_optimize_grid()
    return

######## Execution ########
if __name__ == "__main__":
    main()



#! /usr/env/bin python3
'''Fit samples using a bounded multivariate normal model with a mesh
'''
######## Functions ########

def samples_from_catalog(
                         fname_catalog,
                         event,
                         group,
                         coord_tag,
                        ):
    '''load Gravitational Wave samples from a catalog database
    
    Parameters
    ----------
    fname_catalog: str
        Input location of hdf5 file where samples are stored
    event: str
        Input GW name for gravitational wave event
    group: str
        Input approximant
    coord_tag: str
        Input coordinate group label
    '''
    import numpy as np
    from gwalk.catalog import Catalog
    from gwalk.catalog.coordinates import coord_tags
    # Identify the variable names belonging to our model
    names = coord_tags[coord_tag]
    # Initialize a catalog pointer
    cata = Catalog(fname_catalog)
    print(fname_catalog)
    print(cata.db.list_items())
    # Check that our group exists
    assert cata.group_status(event, group, names)
    # Load the samples as a dictionary
    sample_dict = cata.load_data(event,group,names)
    # Create a numpy array of the samples
    samples = []
    for item in sample_dict.keys():
        if item.startswith("prior"):
            # Extract the prior weights
            prior = np.asarray(sample_dict[item])
        else:
            # Append a variable
            samples.append(sample_dict[item])
    # Make the numpy array usable
    samples = np.asarray(samples).T
    # Invert prior
    inv_prior = prior**-1
    return samples, inv_prior

def model_guesses(
                  fname_nal,
                  event,
                  X, scale, coord_tag
                 ):
    '''Return model guesses for event
    fname_nal: str
        Input location of hdf5 file where fits are stored
    event: str
        Input GW name for gravitational wave event
    X: array like, shape = (nparam)
        Input some guess of fit parameters
    scale: array like, shape = (ndim)
        Input scale lengths for each variable
    coord_tag: str
        Input coordinate group label
    '''
    from gwalk.data import Database
    from gwalk.bounded_multivariate_normal import MultivariateNormal
    from gwalk.catalog.coordinates import coord_tags
    from os.path import join
    import numpy as np
    # Initialize ndim
    ndim = scale.size
    # Load coordinates
    coords = coord_tags[coord_tag][:-1]
    # Open the database
    db = Database(fname_nal,group=event)
    # Get all labels for event
    labels = db.list_items()
    # Initialize guess list
    Xg = []
    # Loop through the labels!
    for item in labels:
        # try to load the fit
        try:
            MVi = MultivariateNormal.load(fname_nal,join(event,item))
        except:
            continue
        # Identify label information
        item_coord_tag, item_group, item_fit_method = item.split(":")
        # Load label coordinates
        item_coords = coord_tags[item_coord_tag][:-1]
        # Create a parameter map
        p_map = {}
        for j, jtem in enumerate(coords):
            p_map[j] = None
            for k, ktem in enumerate(item_coords):
                if jtem == ktem:
                    p_map[j] = k

        # Success! Identify label params
        item_guess = MVi.read_guess()
        item_scale = MVi.scale
        # Initialize new guess
        item_X = X.copy()
        # Loop through each coordinate again
        for j in range(ndim):
            # Do nothing if coordinate isn't in Visitor's coordinate system
            if p_map[j] is None:
                continue
            # Update mu and sigma parameters
            item_X[j] = item_guess[j] * item_scale[p_map[j]]/scale[j]
            item_X[j+ndim] = item_guess[j+ndim] * item_scale[p_map[j]]/scale[j]
            # Update correlation parameters
            for k in range(j):
                if p_map[k] is None:
                    continue
                #Update factor
                item_X[2*ndim + ((j*(j-1))//2) + k] = \
                item_guess[2*ndim+((p_map[j]*(p_map[j]-1))//2)+p_map[k]]

        # Append to guess list
        Xg.append(item_X.copy())
    Xg = np.asarray(Xg)
    #Return guesses!
    return Xg


def kl_div_optimization(mesh, MV):
    '''Return kl divergence based optimization functions

    Parameters
    ----------
    mesh: Mesh object
        Input mesh for sample evaluations
    MV: MultivariateNormal object
        Input truncated Gaussian we are fitting
    '''
    import numpy as np
    def f_opt(X):
        L = np.zeros(X.shape[0])
        k = MV.satisfies_constraints(X)
        L[k] = np.power(mesh.nal_kl_div(MV,X=X[k]),-1)
        return L

    def f_opt_param(X):
        L = np.zeros(X.shape)
        k = MV.satisfies_constraints(X)
        L[k] = np.power(mesh.nal_kl_div(MV,X=X[k],mode='parameter'),-1)
        #print(np.std(L,axis=0))
        return L
    return f_opt, f_opt_param

def kl_diff(
            fname_nal,
            label_simple,
            label_select,
            samples,
            weights=None,
            evaluation_res = 10,
            nwalk=100,
            nstep=100,
            sig_factor=1.0,
            carryover=0.03,
            p_labels=None,
            attrs=None,
            event=None,
            coord_tag=None,
            verbose=False,
            **mesh_kwargs
           ):
    '''Fit samples to a truncated gaussian
    
    Note: this does not need to be a Gravitational Wave event

    Parameters
    ----------
    fname_nal: str
        Input name of the file where fits will be saved
    label: str
        Input label for saving fit within file
    samples: array like, shape = (npts,ndim)
        Input samples for fitting a truncated Gaussian to
    weights: array like, shape = (npts), optional
        Input weights for each sample
    evaluation_res: int, optional
        Input size of marginal evaluations on 1D and 2D marginals
    nwalk: int, optional
        Input number of random walkers to evolve
    nstep: int, optional
        Input number of steps for random walkers
    sig_factor: float, optional
        Input related to jump size for random walkers.
            Don't touch this unless you know what you are doing
    carryover: float, optional
        Input controls fraction of carryover for genetic algorithm
    p_labels: list, optional
        Input parameter labels for fits
    attrs: dict, optional
        Input attrs to save with fits
    event: str, optional
        Input If this is a Gravitational Wave event, we can make some smarter
            guesses in the beginning
    coord_tag: str, optional
        Input If this is a Gravitational Wave event,
            we may want information about the coordinates
    verbose: bool, optional
        Input print things
    '''
    from gwalk.density import Mesh
    from gwalk.bounded_multivariate_normal import MultivariateNormal
    import numpy as np
    import time
    # Identify fit
    # Identify ndim
    ndim = samples.shape[1]
    # Fit the mesh
    mesh = Mesh.fit(
                    samples,
                    ndim,
                    weights=weights,
                    verbose=verbose,
                    **mesh_kwargs
                   )

    # Generate an evaluation set
    mesh.generate_evaluation_set(evaluation_res)
    # Generate a multivariate normal object
    MV_simple = MultivariateNormal.load(fname_nal,label_simple)
    MV_select = MultivariateNormal.load(fname_nal,label_select)
    kl_simple = mesh.nal_kl_div(MV_simple)
    kl_select = mesh.nal_kl_div(MV_select)
    #print(MV_simple.scale)
    #print(MV_select.scale)
    #raise Exception
    print("simple fit\n  kl: %f\tparams:%s"%(kl_simple,str(MV_simple.read_guess())))
    print("select fit\n  kl: %f\tparams:%s"%(kl_select,str(MV_select.read_guess())))
    print("diff:")
    print(np.abs(MV_simple.read_guess() - MV_select.read_guess()))
    print("analytic kl difference:", MV_select.analytic_kl(MV_simple))


def kl_diff_real_event(
                   fname_catalog,
                   fname_nal,
                   event,
                   group,
                   coord_tag,
                   random_state=None,
                   **kwargs
                  ):
    '''\
    Fit some samples to a mesh
    Parameters
    ----------
    fname_catalog: str
        Input location of hdf5 file where samples are stored
    fname_nal: str
        Input name of the file where fits will be saved
    event: str
        Input GW name for gravitational wave event
    group: str
        Input approximant
    coord_tag: str
        Input coordinate group label
    '''
    from gwalk.catalog.coordinates import coord_tags, coord_labels
    from gwalk.density import Mesh
    import numpy as np
    import time
    # Identify fit
    mesh_label = "%s/%s:%s"%(event,coord_tag,group)
    nal_label_simple = "%s/%s:%s:%s"%(event,coord_tag,group,"simple")
    nal_label_select = "%s/%s:%s:%s"%(event,coord_tag,group,"select")

    # load samples
    samples, inv_prior = \
        samples_from_catalog(fname_catalog,event,group,coord_tag)

    # Identify ndim
    ndim = samples.shape[1]

    # Identify coordinate labels
    p_labels = []
    for i in range(ndim):
        p_labels.append(coord_labels[coord_tags[coord_tag][i]])

    # Identify attributes
    attrs = {
             "event"    : event,
             "coord_tag": coord_tag,
             "coords"   : coord_tags[coord_tag],
             "group"    : group,
            }


    '''
    # Determine if mesh can be loaded
    if Mesh.exists(fname_mesh,label=mesh_label):
        mesh = Mesh.load(fname_mesh,label=mesh_label)
    else:
        # Fit the mesh
        mesh = Mesh.fit(
                        samples,
                        ndim,
                        weights=inv_prior,
                        **mesh_kwargs
                       )

        # Save the mesh
        mesh.save(fname_mesh,label=mesh_label)
    '''

    # Generate mesh guesses
    kl_diff(
                      fname_nal,
                      nal_label_simple,
                      nal_label_select,
                      samples,
                      weights=inv_prior,
                      p_labels=p_labels,
                      attrs=attrs,
                      event=event,
                      coord_tag=coord_tag,
                      **kwargs
                     )



######## Main ########
def main():
    import time
    import sys
    fname_catalog = sys.argv[1]
    fname_mesh = sys.argv[2]
    fname_nal = sys.argv[3]
    event = sys.argv[4]
    group = sys.argv[5]
    coord_tag = sys.argv[6]
    min_bins = int(sys.argv[7])
    max_bins = int(sys.argv[8])
    verbose = True
    kl_diff_real_event(
                   fname_catalog,
                   fname_nal,
                   event,
                   group,
                   coord_tag,
                   min_bins=min_bins,
                   max_bins1d=max_bins,
                   max_bins2d=max_bins,
                   whitenoise=0.001,
                   evaluation_res=100,
                   nwalk=100,
                   nstep=100,
                   sig_factor=0.5,
                   carryover=0.03,
                   verbose=True,
                  )
    return

######## Execution ########
if __name__ == "__main__":
    main()



'''Python evaluation of C extensions for multivariate normal distributions'''

######## Imports ########
import numpy as np

######## Functions ########
## dimension checking/reduction ##
def nparam_of_ndim(ndim):
    '''Return the number of parameters associated witha guess in n dimensions
    Parameters
    ----------
    ndim: int
        Input number of dimensions
    '''
    return int((ndim*ndim + 3*ndim)//2) + 1

def ndim_of_nparam(nparam):
    '''find the dimensionality of a guess
    Parameters
    ----------
    nparam: int
        Input number of parameters
    Returns
    -------
    ndim: int
        Number of dimensions
    '''
    for i in range(nparam):
        if nparam == nparam_of_ndim(i):
            return i
    raise RuntimeError("Failed to get number of dimensions for guess")

def offset_of_params(X):
    '''Reconstruct the normalization constant from parameters
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    Returns
    -------
    offset: `~numpy.ndarray` shape = (ngauss, 1)
        Output array of parameter offsets
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    norm = X[:,0].copy()
    return norm

def mu_of_params(X):
    '''Reconstruct the vectorized mu parameters of input Gaussian parameters

    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    Returns
    -------
    mu: `~numpy.ndarray` shape = (ngauss, ndim)
        Output array of mu values
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Identify n_gauss
    n_gauss = X.shape[0]
    # Identify ndim
    nparam = X.shape[1]
    ndim = ndim_of_nparam(nparam)
    # Reconstruct mu
    mu = np.copy(X[:,1:ndim+1])
    return mu
    
def std_of_params(X):
    '''Reconstruct the vectorized sigma parameters of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    Returns
    -------
    std: `~numpy.ndarray` shape = (ngauss, ndim)
        Output array of std values
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Identify n_gauss
    n_gauss = X.shape[0]
    # Identify ndim
    nparam = X.shape[1]
    ndim = ndim_of_nparam(nparam)
    # Reconstruct std
    std = np.copy(X[:,ndim+1:2*ndim+1])
    return std

def cor_of_params(X):
    '''Reconstruct the vectorized correlation matrix of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    Returns
    -------
    cor: `~numpy.ndarray` shape = (ngauss, ndim, ndim)
        Output array of correlation matrices
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Identify n_gauss
    n_gauss = X.shape[0]
    # Identify ndim
    nparam = X.shape[1]
    ndim = ndim_of_nparam(nparam)
    # Calculate std of params
    std = std_of_params(X)
    # Reconstruct correlation matrices
    cor = np.ones((n_gauss, ndim, ndim))*np.eye(ndim)
    # The correlation index is carefully kept consistent
    # DO NOT CHANGE THIS!
    cor_index = 2*ndim + 1
    for i in range(ndim):
        for j in range(i):
            cor[:,i,j] = cor[:,j,i] = X[:,cor_index]
            cor_index += 1
    if not cor_index == nparam:
        raise RuntimeError("Correlation Matrix reconstruction is broken")
    return cor

def cov_of_params(X,**kwargs):
    '''Reconstruct the vectorized corvariance of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Identify n_gauss
    n_gauss = X.shape[0]
    # Identify ndim
    nparam = X.shape[1]
    ndim = ndim_of_nparam(nparam)
    # Calculate std of params
    std = std_of_params(X,**kwargs)
    # Calculate correlation matrix
    cor = cor_of_params(X,**kwargs)
    # Reconstruct the variance matrix
    # DO NOT CHANGE THIS!
    std_expand = np.tensordot(std, np.ones(ndim), axes=0)
    var = std_expand * np.transpose(std_expand, axes=[0,2,1])
    # Reconstruct the covariance matrix
    cov = cor * var
    return cov

def mu_cov_of_params(X, **kwargs):
    '''Reconstruct vectorized mu and covariance of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    '''
    mu = mu_of_params(X,**kwargs)
    cov = cov_of_params(X,**kwargs)
    return mu, cov

def offset_mu_cov_of_params(X, **kwargs):
    '''Reconstruct vectorized mu and covariance of input Gaussian parameters
    Parameters
    ----------
    X: array like, shape = (ngauss, nparams)
        Input Array of parameter guesses
    '''
    mu = mu_of_params(X, **kwargs)
    cov = cov_of_params(X, **kwargs)
    offset = offset_of_params(X, **kwargs)
    return offset, mu, cov

def cov_of_std_cor(std,cor):
    '''Reconstruct the covariance from std and corelation
    Parameters
    ----------
    std: array like, shape = (npts, ndim)
        Input array of sigma values
    cor: array like, shape = (npts, ndim, ndim)
        Input array of correlation matrix values
    '''
    # Check dimensionality of std
    if len(std.shape) == 1:
        npts, ndim = 1, std.size
        std = std.reshape((npts,ndim))
    else:
        npts, ndim = std.shape
    # Check dimensionality of cor
    if len(cor.shape) == 2:
        cor = cor.reshape((npts,ndim,ndim))
    # Assert dimensionality match up
    assert cor.shape == ((npts,ndim,ndim))

    # DO NOT CHANGE THIS!
    std_expand = np.tensordot(std, np.ones(ndim), axes=0)
    var = std_expand * np.transpose(std_expand, axes=[0,2,1])
    # Reconstruct the covariance matrix
    cov = cor * var
    return cov

def std_of_cov(cov):
    '''Extract the vectorized std from the vectorized covariance
    Parameters
    ----------
    cov: array like, shape = (npts, ndim, ndim)
        Input Array of cov values
    '''
    # Protect against usage
    if len(cov.shape) == 2:
        cov = cov[None,:,:]
    elif len(cov.shape) != 3:
        raise ValueError("Error std_of_cov")
    # Check squareness
    ndim = cov.shape[-1]
    if not (cov.shape[-2] == ndim):
        raise ValueError("cov must be square")
    # Identify std values
    std = np.sqrt(np.sum(cov*np.eye(ndim),axis=1))
    return std

def cor_of_cov(cov):
    '''Reconstruct the vectorized correlation matrix from the covariance
    Parameters
    ----------
    cov: array like, shape = (npts, ndim, ndim)
        Input Array of cov values
    '''
    # Protect against usage
    if len(cov.shape) == 2:
        cov = cov[None,:,:]
    elif len(cov.shape) != 3:
        raise ValueError("Error cor_of_cov")
    # Check squareness
    ndim = cov.shape[-1]
    if not (cov.shape[-2] == ndim):
        raise ValueError("cov must be square")
    # Identify std values
    std = np.sqrt(np.sum(cov*np.eye(ndim),axis=1))
    # Compute the inverse variance
    istd = 1./std
    # Compute the expansion of istd
    istd_expand = np.tensordot(istd, np.ones(ndim), axes=0)
    # Compute the inverse of the fully correlated covariance
    isig2 = istd_expand * np.transpose(istd_expand, axes=[0,2,1])
    # Calculate the correlation matrix
    cor = cov*isig2
    return cor

#### Functions that assume normal model object ####
def params_of_offset_mu_cov(norm, mu, cov):
    '''Reconstruct Gaussian vectorized parameters for given mu and covariance
    Parameters
    ----------
    mu: array like, shape = (npts, ndim)
        Input Array of mu values
    cov: array like, shape = (npts, ndim, ndim)
        Input Array of cov values
    '''
    # Protect against usage
    if (len(mu.shape) == 1) and (len(cov.shape) == 2):
        mu = mu[None,:]
        cov = cov[None,:,:]
    elif (len(mu.shape) != 2) or (len(cov.shape) != 3):
        raise ValueError("mu or cov is not the right shape!")
    # Identify useful information
    n_gauss = mu.shape[0]
    ndim = mu.shape[1]
    nparam = nparam_of_ndim(ndim)
    # Protect against usage
    if cov.shape != (n_gauss, ndim, ndim):
        raise ValueError("cov is not shape (n_gauss, ndim, ndim)")
    # Initialize array
    X = np.empty((n_gauss,nparam))
    # Initialize normalization
    X[:,0] = norm
    # Insert mu values
    X[:,1:ndim+1] = mu
    # Insert std values
    X[:,ndim+1:2*ndim+1] = np.sqrt(np.sum(cov*np.eye(ndim),axis=1))
    ## Compute Correlation ##
    cor = cor_of_cov(cov)
    ## Insert Correlation Values ##
    cor_index = 2*ndim +1
    for i in range(ndim):
        for j in range(i):
            X[:,cor_index] = cor[:,i,j]
            cor_index += 1
    if not cor_index == nparam:
        raise RuntimeError("Correlation matrix not extracted properly")
    # And we're done!
    return X

def params_of_mu_cov(mu, cov):
    '''Reconstruct Gaussian vectorized parameters for given mu and covariance
    Parameters
    ----------
    mu: array like, shape = (npts, ndim)
        Input Array of mu values
    cov: array like, shape = (npts, ndim, ndim)
        Input Array of cov values
    '''
    # Protect against usage
    if (len(mu.shape) == 1) and (len(cov.shape) == 2):
        mu = mu[None,:]
        cov = cov[None,:,:]
    elif (len(mu.shape) != 2) or (len(cov.shape) != 3):
        raise ValueError("mu or cov is not the right shape!")
    # Identify useful information
    n_gauss = mu.shape[0]
    ndim = mu.shape[1]
    nparam = nparam_of_ndim(ndim)
    # Protect against usage
    if cov.shape != (n_gauss, ndim, ndim):
        raise ValueError("cov is not shape (n_gauss, ndim, ndim)")
    # Initialize array
    X = np.empty((n_gauss,nparam))
    # Initialize normalization
    X[:,0] = 0.0
    # Insert mu values
    X[:,1:ndim+1] = mu
    # Insert std values
    X[:,ndim+1:2*ndim+1] = np.sqrt(np.sum(cov*np.eye(ndim),axis=1))
    ## Compute Correlation ##
    cor = cor_of_cov(cov)
    ## Insert Correlation Values ##
    cor_index = 2*ndim +1
    for i in range(ndim):
        for j in range(i):
            X[:,cor_index] = cor[:,i,j]
            cor_index += 1
    if not cor_index == nparam:
        raise RuntimeError("Correlation matrix not extracted properly")
    # And we're done!
    return X

def params_reduce_dd(X, indices):
    '''Reduce to indexed marginals
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    indices: list
        Input list of indices we would like parameters for
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Extract dimensionality
    n_gauss, nparam = X.shape
    ndim = ndim_of_nparam(nparam)
    # Extract dimensionality of indices
    ndim_index = len(indices)
    nparam_index = nparam_of_ndim(ndim_index)
    # Initialize array
    Xp = np.empty((n_gauss, nparam_index))
    # Set normalization
    Xp[:,0] = X[:,0]
    # Loop through Xp dimensions
    for i in range(ndim_index):
         Xp[:,i+1] =              X[:,indices[i] + 1]
         Xp[:,i+ndim_index+1] =   X[:,indices[i]+ndim + 1]
    # Loop through correlation factors
    for i in range(ndim_index):
        for j in range(i):
            if indices[i] > indices[j]:
                index, jndex = indices[i], indices[j]
            elif indices[i] < indices[j]:
                index, jndex = indices[j], indices[i]
            else:
                raise RuntimeError("indices[i] == indices[j]")
            Xp[:,2*ndim_index + ((i*(i-1))//2) + j + 1] = \
                X[:,2*ndim + ((index*(index-1))//2) + jndex + 1]
    return Xp

def params_reduce_1d(X, index):
    '''Reduce to a 1D marginal set of parameters
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    index: int
        Input which dimension would we like to generate an evaluation set for
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Extract dimensionality
    n_gauss, nparam = X.shape
    ndim = ndim_of_nparam(nparam)
    # Protect against invalid choices
    assert ndim > index
    # Reduce dimensionality
    Xi = np.asarray([X[:,0], X[:,index + 1],X[:,index+ndim + 1]]).T
    return Xi


######## Modify Line ########


def params_reduce_2d(X, index, jndex):
    '''Reduce to a 2D marginal set of parameters
    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input Array of parameter guesses
    index: int
        Input which dimension would we like to generate an evaluation set for
    jndex: int
        Input which other dimension would we like to generate an evaluation set for
    '''
    # Protect against single set of parameters
    if len(X.shape) == 1:
        X = X[None,:]
    # Extract dimensionality
    n_gauss, nparam = X.shape
    ndim = ndim_of_nparam(nparam)
    # Protect usage
    assert ndim > index
    assert index > jndex
    # Reduce dimensionality
    Xij = np.asarray([
                      X[:,0],
                      X[:,index + 1],
                      X[:,jndex + 1],
                      X[:,index+ndim + 1],
                      X[:,jndex+ndim + 1],
                      X[:,2*ndim + ((index*(index-1))//2) + jndex + 1]
                     ]).T
    return Xij




#!/home/xevra/.local/bin/python3
'''Reformulate package catalog samples'''

######## Imports ########
import numpy as np
import os
from os.path import join, isdir, isfile
from xdata import Database
from gwalk.catalog.coordinates import coord_tags as COORD_TAGS
from gwalk.catalog.coordinates import coord_labels as COORD_LABELS
from gwalk.catalog.coordinates import q_of_mc_eta
from gwalk.catalog.prior.prior_methods import prior_mc_eta as prior_mass
from gwalk.catalog.prior.prior_methods import prior_dist
from gwalk.catalog.prior.callister_prior import chi_effective_prior_of_aligned_spins as chi_eff_aligned_prior
from gwalk.catalog.prior.prior_methods import prior_full_spin


######## Globals ########


EVENTS = {
          "GWTC-1" : [
                      "GW150914",
                      "GW151012",
                      "GW151226",
                      "GW170104",
                      "GW170608",
                      "GW170729",
                      "GW170809",
                      "GW170814",
                      "GW170817",
                      "GW170818",
                      "GW170823",
                     ],
            "GWTC-2" : [
                        'GW190408_181802',
                        'GW190412',
                        'GW190413_052954',
                        'GW190413_134308',
                        'GW190421_213856',
                        'GW190424_180648',
                        'GW190425',
                        'GW190426_152155',
                        'GW190503_185404',
                        'GW190512_180714',
                        'GW190513_205428',
                        'GW190514_065416',
                        'GW190517_055101',
                        'GW190519_153544',
                        'GW190521_074359',
                        'GW190521',
                        'GW190527_092055',
                        'GW190602_175927',
                        'GW190620_030421',
                        'GW190630_185205',
                        'GW190701_203306',
                        'GW190706_222641',
                        'GW190707_093326',
                        'GW190708_232457',
                        'GW190719_215514',
                        'GW190720_000836',
                        'GW190727_060333',
                        'GW190728_064510',
                        'GW190731_140936',
                        'GW190803_022701',
                        'GW190814',
                        'GW190828_063405',
                        'GW190828_065509',
                        'GW190909_114149',
                        'GW190910_112807',
                        'GW190915_235702',
                        'GW190924_021846',
                        'GW190929_012149',
                        'GW190930_133541',
                       ],
            "GWTC-2p1" : [
                          "GW190403_051519",
                          "GW190426_190642",
                          "GW190725_174728",
                          "GW190805_211137",
                          "GW190916_200658",
                          "GW190917_114630",
                          "GW190925_232845",
                          "GW190926_050336",
                         ],
            "GWTC-3" : [
                        "GW191103_012549",
                        "GW191105_143521",
                        "GW191109_010717",
                        "GW191113_071753",
                        "GW191126_115259",
                        "GW191127_050227",
                        "GW191129_134029",
                        "GW191204_110529",
                        "GW191204_171526",
                        "GW191215_223052",
                        "GW191216_213338",
                        "GW191219_163120",
                        "GW191222_033537",
                        "GW191230_180458",
                        "GW200105_162426",
                        "GW200112_155838",
                        "GW200115_042309",
                        "GW200128_022011",
                        "GW200129_065458",
                        "GW200202_154313",
                        "GW200208_130117",
                        "GW200208_222617",
                        "GW200209_085452",
                        "GW200210_092255",
                        "GW200216_220804",
                        "GW200219_094415",
                        "GW200220_061928",
                        "GW200220_124850",
                        "GW200224_222234",
                        "GW200225_060421",
                        "GW200302_015811",
                        "GW200306_093714",
                        "GW200308_173609",
                        "GW200311_115853",
                        "GW200316_215756",
                        "GW200322_091133",
                       ],
        }

EXTENSION = {
             "GWTC-1"   : "_GWTC-1.hdf5",
             "GWTC-2"   : ".h5",
             "GWTC-2p1" : ".h5",
             "GWTC-3"   : ".h5",
            }

GROUP_ALIAS = {
               "IMRPhenomPv2_posterior"                     : "IMRPhenomPv2",
               "SEOBNRv3_posterior"                         : "SEOBNRv3",
               "Overall_posterior"                          : "Overall",
               "IMRPhenomPv2NRT_highSpin_posterior"         : "IMRPhenomPv2NRT_highSpin",
               "IMRPhenomPv2NRT_lowSpin_posterior"          : "IMRPhenomPv2NRT_lowSpin",
               "C01:IMRPhenomD"                             : "IMRPhenomD",
               "C01:IMRPhenomPv2"                           : "IMRPhenomPv2",
               "C01:SEOBNRv4P"                              : "SEOBNRv4P",
               "C01:SEOBNRv4PHM"                            : "SEOBNRv4PHM",
               "C01:IMRPhenomHM"                            : "IMRPhenomHM",
               "C01:IMRPhenomPv3HM"                         : "IMRPhenomPv3HM",
               "C01:SEOBNRv4HM_ROM"                         : "SEOBNRv4HM_ROM",
               "C01:SEOBNRv4_ROM"                           : "SEOBNRv4_ROM",
               "C01:NRSur7dq4"                              : "NRSur7dq4",
               "C01:SEOBNRv4P_nonevol"                      : "SEOBNRv4P_nonevol",
               "C01:IMRPhenomD_NRTidal-HS"                  : "IMRPhenomD_NRTidal-HS",
               "C01:IMRPhenomD_NRTidal-LS"                  : "IMRPhenomD_NRTidal-LS",
               "C01:IMRPhenomPv2_NRTidal-HS"                : "IMRPhenomPv2_NRTidal-HS",
               "C01:IMRPhenomPv2_NRTidal-LS"                : "IMRPhenomPv2_NRTidal-LS",
               "C01:SEOBNRv4T_surrogate_HS"                 : "SEOBNRv4T_surrogate_HS",
               "C01:SEOBNRv4T_surrogate_LS"                 : "SEOBNRv4T_surrogate_LS",
               "C01:SEOBNRv4T_surrogate_highspin_RIFT"      : "SEOBNRv4T_surrogate_highspin_RIFT",
               "C01:SEOBNRv4T_surrogate_lowspin_RIFT"       : "SEOBNRv4T_surrogate_lowspin_RIFT",
               "C01:TEOBResumS-HS"                          : "TEOBResumS-HS",
               "C01:TEOBResumS-LS"                          : "TEOBResumS-LS",
               "C01:TaylorF2-HS"                            : "TaylorF2-HS",
               "C01:TaylorF2-LS"                            : "TaylorF2-LS",
               "C01:IMRPhenomNSBH"                          : "IMRPhenomNSBH",
               "C01:TaylorF2"                               : "TaylorF2",
               "C01:SEOBNRv4P_RIFT"                         : "SEOBNRv4P_RIFT",
               "C01:IMRPhenomXPHM"                          : "IMRPhenomXPHM",
               "C01:Mixed"                                  : "Mixed",
               "C01:IMRPhenomNSBH:HighSpin"                 : "IMRPhenomNSBH_HighSpin",
               "C01:IMRPhenomNSBH:LowSpin"                  : "IMRPhenomNSBH_LowSpin",
               "C01:IMRPhenomXPHM:HighSpin"                 : "IMRPhenomXPHM_HighSpin",
               "C01:IMRPhenomXPHM:LowSpin"                  : "IMRPhenomXPHM_LowSpin",
               "C01:Mixed:NSBH:HighSpin"                    : "Mixed_NSBH_HighSpin",
               "C01:Mixed:NSBH:LowSpin"                     : "Mixed_NSBH_LowSpin",
               "C01:SEOBNRv4_ROM_NRTidalv2_NSBH"            : "SEOBNRv4_ROM_NRTidalv2_NSBH",
               "C01:SEOBNRv4_ROM_NRTidalv2_NSBH:HighSpin"   : "SEOBNRv4_ROM_NRTidalv2_NSBH_HighSpin",
               "C01:SEOBNRv4_ROM_NRTidalv2_NSBH:LowSpin"    : "SEOBNRv4_ROM_NRTidalv2_NSBH_LowSpin",
              }

######## Functions ########

#### Read Samples ####

def event_groups(fname_event):
    '''find all the groups for an event in a catalog file
    Parameters
    ----------
    fname_event: str
        Input the file location for posterior samples
    '''
    # Check that the database exists
    assert isfile(fname_event)
    # Open it up
    db = Database(fname_event)
    # Initialize the list of groups
    groups = []
    # Loop through each group
    for item in db.list_items('.'):
        if db.kind(item) == 'group':
            # Check if there are posterior samples for that group
            if "posterior_samples" in db.list_items(item,kind='dset'):
                # if there are, it's safe to say it's a waveform
                groups.append(item)
        elif item.endswith("posterior"):
            groups.append(item)
    return groups

def load_GWTC_1_event_samples(fname_event, group):
    ''' Load GWTC_1 Event samples 
    Parameters
    ----------
    fname_event: str
        Input the file location for posterior samples
    group: str
        Input the label for the waveform model posterior samples
    '''
    from .read_catalog_GWTC_1 import generate_samples
    # Generate samples
    pdict = generate_samples(fname_event, group)
    return pdict

def load_GWTC_2_event_samples(fname_event, group):
    ''' Load GWTC_2 Event samples 
    Parameters
    ----------
    fname_event: str
        Input the file location for posterior samples
    group: str
        Input the label for the waveform model posterior samples
    '''
    from .read_catalog_GWTC_2 import generate_samples
    # Generate samples
    pdict = generate_samples(fname_event, group)
    return pdict

def load_GWTC_2p1_event_samples(fname_event, group):
    ''' Load GWTC_3 Event samples 
    Parameters
    ----------
    fname_event: str
        Input the file location for posterior samples
    group: str
        Input the label for the waveform model posterior samples
    '''
    from .read_catalog_GWTC_2p1 import generate_samples
    # Generate samples
    pdict = generate_samples(fname_event, group)
    return pdict

def load_GWTC_3_event_samples(fname_event, group):
    ''' Load GWTC_3 Event samples 
    Parameters
    ----------
    fname_event: str
        Input the file location for posterior samples
    group: str
        Input the label for the waveform model posterior samples
    '''
    from .read_catalog_GWTC_3 import generate_samples
    # Generate samples
    pdict = generate_samples(fname_event, group)
    return pdict

def load_event_samples(release, fname_event, group):
    '''Load Release Samples 
    Parameters
    ----------
    release: str
        Input the release associated with the event
    fname_event: str
        Input the file location for posterior samples
    group: str
        Input the label for the waveform model posterior samples
    '''
    # Generate samples
    if release == "o3a":
        pdict = load_o3a_event_samples(fname_event, group)
    elif release == "GWTC-1":
        pdict = load_GWTC_1_event_samples(fname_event, group)
    elif release == "GWTC-2":
        pdict = load_GWTC_2_event_samples(fname_event, group)
    elif release == "GWTC-2p1":
        pdict = load_GWTC_2p1_event_samples(fname_event, group)
    elif release == "GWTC-3":
        pdict = load_GWTC_3_event_samples(fname_event, group)
    else:
        raise ValueError("Unknown sample release")
    return pdict



######## CATALOG Object ########

class Catalog(object):
    '''Handle pointers to the catalog database file'''
    def __init__(
                 self,
                 fname,
                 clean=False,
                ):
        '''Initialize catalog object
        Parameters
        ----------
        fname: str
            Input name of catalog file
        clean: bool
            Delete the database and make a new one?
        '''
        self.fname = fname
        if clean and isfile(fname):
            os.system("rm %s"%fname)
        self.db = Database(fname,sleep=0)

    def events_of_release(self, release):
        '''Return events associated with a given release
        Parameters
        ----------
        release: str
            Input A label for a Gravitational-Wave Tranient Catalog
        '''
        return EVENTS[release]

    def events_of_group(self, group):
        '''Return a list of all the events which have a particular group
        Parameters
        ----------
        group: str
            Input waveform identifier
        '''
        events = []
        for release in EVENTS:
            for event in EVENTS[release]:
                if group in self.groups_of_event(event):
                    events.append(event)
        return events

    def release_of_event(self, event):
        '''Return the release associated with an event
        Parameters
        ----------
        event: str
            Input the GW tag for an event in the GWTC
        '''
        for release in EVENTS:
            if event in EVENTS[release]:
                return release

    def release_of_group(self, group):
        '''Return the release associated with a group
        Parameters
        ----------
        group: str
            Input waveform identifier
        '''
        releases = []
        for release in EVENTS:
            for event in EVENTS[release]:
                if group in self.groups_of_event(event):
                    releases.append(release)
        return releases

    def groups_of_event(self, event):
        '''Return the groups associated with an event
        Parameters
        ----------
        event: str
            Input the GW tag for an event in the GWTC
        '''
        release = self.release_of_event(event)
        event_addr = join(release, event)
        return self.db.list_items(event_addr, kind='group')

    def groups_of_release(self, release):
        '''Return the groups associated with release events
        Parameters
        ----------
        release: str
            Input A label for a Gravitational-Wave Tranient Catalog
        '''
        groups = []
        events = self.events_of_release(release)
        for event in events:
            for item in self.groups_of_event(event):
                if not item in groups:
                    groups.append(item)
        return groups

    def fields_of_group(self, group, mode='any'):
        '''Return a list of the fields available for a group
        Parameters
        ----------
        group: str
            Input waveform identifier
        mode: str
            Input 'any' or 'all'
        '''
        fields = []
        events = self.events_of_group(group)
        for event in events:
            for field in self.fields_of_event_group(event,group):
                if not field in fields:
                    fields.append(field)
        if mode == "any":
            return fields
        elif mode == "all":
            common = []
            for field in fields:
                has_field = True
                for event in events:
                    if not field in self.fields_of_event_group(event,group):
                        has_field = False
                if has_field:
                    common.append(field)
            return common
        else:
            raise RuntimeError("Unknown mode: %s"%mode)

    def fields_of_event_group(self, event, group):
        '''Return a list of the fields available for a group
        Parameters
        ----------
        event: str
            Input the GW tag for an event in the GWTC
        group: str
            Input waveform identifier
        '''
        release = self.release_of_event(event)
        group_addr = join(release, event, group)
        return self.db.list_items(group_addr)

    def tags_of_event_group(self, event, group):
        '''
        Parameters
        ----------
        event: str
            Input the GW tag for an event in the GWTC
        group: str
            Input waveform identifier
        '''
        fields = self.fields_of_event_group(event, group)
        tags = []
        for tag in COORD_TAGS:
            has_tag = True
            for item in COORD_TAGS[tag]:
                if not (item in fields):
                    has_tag = False
            if has_tag:
                tags.append(tag)

        return tags

    def tags_of_event(self, event):
        '''
        Parameters
        ----------
        event: str
            Input the GW tag for an event in the GWTC
        '''
        groups = self.groups_of_event(event)
        tags = []
        for item in groups:
            group_tags = self.tags_of_event_group(event, item)
            for jtem in group_tags:
                if not jtem in tags:
                    tags.append(jtem)
        return tags

    def groups_of_event_tag(self, event, coord_tag):
        '''
        Parameters
        ----------
        event: str
            Input the GW tag for an event in the GWTC
        coord_tag: str
            Input label for set of coordinates
        '''
        groups = []
        for item in self.groups_of_event(event):
            if coord_tag in self.tags_of_event_group(event, item):
                groups.append(item)
        return groups

                
    def tags_of_group(self, group):
        '''
        Parameters
        ----------
        group: str
            Input waveform identifier
        '''
        fields = self.fields_of_group(group,mode='all')
        tags = []
        for tag in COORD_TAGS:
            has_tag = True
            for item in COORD_TAGS[tag]:
                if not (item in fields):
                    has_tag = False
            if has_tag:
                tags.append(tag)

        return tags

                
    def group_has_tides(self, event, group):
        '''check if a group has tidal parameters
        Parameters
        ----------
        event: str
            Input the GW tag for an event in the GWTC
        group: str
            Input waveform identifier
        '''
        # Identify the release
        release = self.release_of_event(event)
        # Find the address for the group
        group_addr = join(release, event, group)
        # Check for lambda1
        datasets = self.db.list_items(group_addr)
        if "lambda_1" in datasets:
            return True
        else:
            return False

    def event_has_tides(self, event):
        '''check if an event has tidal parameters
        Parameters
        ----------
        event: str
            Input the GW tag for an event in the GWTC
        '''
        # Get groups for event
        groups = self.groups_of_event(event)
        # Initialize truth value
        tides = False
        # Loop through each group
        for group in groups:
            if self.group_has_tides(event, group):
                tides = True
        return tides

    def all_tidal_events(self):
        '''Find all the tidal events
        Parameters
        ----------
        '''
        # Initialize list of tidal events
        tidal_events = []
        # Loop through each release
        for release in EVENTS:
            # Loop through each event
            for event in EVENTS[release]:
                if self.event_has_tides(event):
                    tidal_events.append(event)
        # Print update to user
        print("All tides events: ", tidal_events)

    def group_has_full_spin(self, event, group):
        '''Check if a group has full spin parameters for its posterior
        Parameters
        ----------
        event: str
            Input the GW tag for an event in the GWTC
        group: str
            Input waveform identifier
        '''
        # Identify the release
        release = self.release_of_event(event)
        # Find the address for the group
        group_addr = join(release, event, group)
        # Check for spin
        datasets = self.db.list_items(group_addr)
        if "spin_x" in datasets:
            return True
        else:
            return False

    def event_has_full_spin(self, event):
        '''check if an event has tidal parameters
        Parameters
        ----------
        event: str
            Input the GW tag for an event in the GWTC
        '''
        # Get groups for event
        groups = self.groups_of_event(event)
        # Initialize truth value
        full_spin = False
        # Loop through each group
        for group in groups:
            if self.group_has_full_spin(event, group):
                full_spin = True
        return full_spin

    def full_spin_groups(self, event):
        '''Return a list of groups for an event which have full spin
        Parameters
        ----------
        event: str
            Input the GW tag for an event in the GWTC
        '''
        # Initialize list
        group_list = []
        # Loop through groups
        groups = self.groups_of_event(event)
        for group in groups:
            if self.group_has_full_spin(event, group):
                group_list.append(group)
        # Return groups
        return group_list

    def load_data(
                  self,
                  event,
                  group,
                  names
                 ):
        '''Read data from a group for a waveform
        Parameters
        ----------
        event: str
            Input the GW tag for an event in the GWTC
        group: str
            Input waveform identifier
        names: list
            Input names of coordinates
        '''
        # Identify the release
        release = self.release_of_event(event)
        # Find the address for the group
        group_addr = join(release, event, group)
        # Initialize the dictionary
        pdict = {}
        # Loop through each name
        for item in names:
            # Load data
            pdict[item] = self.db.dset_value(join(group_addr, item))
        return pdict

    def group_status(
                     self,
                     event,
                     group,
                     names
                    ):
        '''Check if the group is ready for analysis
        Parameters
        ----------
        event: str
            Input the GW tag for an event in the GWTC
        group: str
            Input waveform identifier
        names: list
            Input names of coordinates
        '''
        # Identify the release
        release = self.release_of_event(event)
        # Find the address for the group
        group_addr = join(release, event, group)
        # Initialize the dictionary
        pdict = {}
        # Loop through each name
        ready = True
        for item in names:
            ready = ready and self.db.exists(join(group_addr, item))
        return ready
    
    def build_prior(
                    self,
                    release,
                    event,
                    group,
                    compression='gzip',
                    spin_max = 1.,
                    mc_min_hard=0.0,
                    mc_max_hard=100.,
                    eta_min_hard=1e-6,
                    eta_max_hard=0.25,
                   ):
        '''Build the prior functions
        Parameters
        ----------
        release: str
            Input the release associated with the event
        event: str
            Input the GW tag for an event in the GWTC
        group: str
            Input waveform identifier
        compression: str, optional
            Input compression level for database
        spin_max: float, optional
            Input maximum allowed spin
        mc_min_hard: float
            Input minimum value for mc
        mc_max_hard: float
            Input maximum value for mc
        eta_min_hard: float
            Input minimum value for eta
        eta_max_hard: float
            Input maximum value for eta
        '''
        # Find the group we are interested in
        group_addr = join(release,event,group)
        # Load coordinates
        mc =        self.db.dset_value(join(group_addr, "chirp_mass"))
        eta =       self.db.dset_value(join(group_addr, "symmetric_mass_ratio"))
        lum_dist =  self.db.dset_value(join(group_addr, "luminosity_distance"))
        chieff =    self.db.dset_value(join(group_addr, "chi_eff"))
        z =         self.db.dset_value(join(group_addr, "redshift"))
        q =         self.db.dset_value(join(group_addr, "mass_ratio"))
        # Compute priors
        mass_prior = prior_mass(
                                mc, eta,
                                mc_min=mc_min_hard, mc_max=mc_max_hard,
                                eta_min=eta_min_hard,eta_max=eta_max_hard, 
                               )
        chieff_aligned_prior = chi_eff_aligned_prior(q, spin_max, chieff)
        aligned_3d_prior = mass_prior*chieff_aligned_prior
        dist_prior = prior_dist(lum_dist)
        aligned_3d_dist_prior = dist_prior*aligned_3d_prior
        # Save priors
        self.db.dset_set(join(group_addr, "prior_mass"), mass_prior, compression=compression)
        self.db.dset_set(join(group_addr, "prior_chi_eff_aligned"), chieff_aligned_prior, compression=compression)
        self.db.dset_set(join(group_addr, "prior_aligned3d"), aligned_3d_prior, compression=compression)
        self.db.dset_set(join(group_addr, "prior_aligned3d_dist"), aligned_3d_dist_prior, compression=compression)
        # Precessing components
        try:
            chi1x = self.db.dset_value(join(group_addr, "spin_1x"))
            chi1y = self.db.dset_value(join(group_addr, "spin_1y"))
            chi1z = self.db.dset_value(join(group_addr, "spin_1z"))
            chi2x = self.db.dset_value(join(group_addr, "spin_2x"))
            chi2y = self.db.dset_value(join(group_addr, "spin_2y"))
            chi2z = self.db.dset_value(join(group_addr, "spin_2z"))
            spin_prior = prior_full_spin(chi1x, chi2x, chi1y, chi2y, chi1z, chi2z)
            precessing8d_prior = mass_prior*spin_prior
            precessing8d_dist_prior = precessing8d_prior*dist_prior
            self.db.dset_set(join(group_addr, "prior_precessing8d"), 
                    precessing8d_prior,compression=compression)
            self.db.dset_set(join(group_addr, "prior_precessing8d_dist"),
                    precessing8d_dist_prior,compression=compression)
        except:
            pass

    def build_catalog(
                      self,
                      release,
                      catalog_directory,
                      compression='gzip',
                      **kwargs
                     ):
        '''build a catalog database
        Parameters
        ----------
        release: str
            Input the release associated with the event
        catalog_directory: str
            Input path to GWXXXXXX.h5 zenodo files
        compression: str, optional
            Input compression level for database
        '''
        print("Building new catalog!")
        # Initialize the release in the database
        if not self.db.exists(release, kind='group'):
            self.db.create_group(release)
        ## Event loop ##
        for event in EVENTS[release]:
            # Identify the event address
            event_addr = join(release, event)
            # Initialize the event in the datanase
            if not self.db.exists(event_addr, kind='group'):
                self.db.create_group(event_addr)
            # Find the event in the catalog directory
            fname_event = join(catalog_directory, event) + EXTENSION[release]
            assert isfile(fname_event)
            # Find the groups in a catalog for a particular event
            groups = event_groups(fname_event)
            # Loop through each group
            for group in groups:
                if group in GROUP_ALIAS:
                    print("alias %s %s"%(group, GROUP_ALIAS[group]))
                    alias = GROUP_ALIAS[group]
                else:
                    print("No alias %s"%(group))
                    alias = group
                # Group address
                group_addr = join(release,event,alias)
                # Initialize the group in the datanase
                if not self.db.exists(group_addr, kind='group'):
                    self.db.create_group(group_addr)
                print("Reading posterior for %s %s"%(event, group))
                # Find the posterior samples in the catalog
                pdict = load_event_samples(release, fname_event, group)
                # Save catalog samples
                for item in pdict:
                    # save the dataset
                    if not self.db.exists(join(group_addr,item), kind='dset'):
                        self.db.dset_set(join(group_addr,item),pdict[item],compression=compression)
                    
                ## Compute the prior ##
                self.build_prior(release,event,alias,compression=compression,**kwargs)
    
